/*
 * linux/fs/ext2/namei.c
 *
 * Rewrite to pagecache. Almost all code had been changed, so blame me
 * if the things go wrong. Please, send bug reports to
 * viro@parcelfarce.linux.theplanet.co.uk
 *
 * Stuff here is basically a glue between the VFS and generic UNIXish
 * filesystem that keeps everything in pagecache. All knowledge of the
 * directory layout is in fs/ext2/dir.c - it turned out to be easily separatable
 * and it's easier to debug that way. In principle we might want to
 * generalize that a bit and turn it into a library. Or not.
 *
 * The only non-static object here is ext2s_dir_inode_operations.
 *
 * TODO: get rid of kmap() use, add readahead.
 *
 * Copyright (C) 1992, 1993, 1994, 1995
 * Remy Card (card@masi.ibp.fr)
 * Laboratoire MASI - Institut Blaise Pascal
 * Universite Pierre et Marie Curie (Paris VI)
 *
 *  from
 *
 *  linux/fs/minix/namei.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *
 *  Big-endian to little-endian byte-swapping/bitmaps by
 *        David S. Miller (davem@caip.rutgers.edu), 1995
 */

#include <linux/pagemap.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/syscalls.h>
#include <asm/unistd.h>
#include <asm/uaccess.h>
#include "ext2.h"
#include "xattr.h"
#include "acl.h"
#include "xip.h"

/*
 * global counter
 */
int ext2s_get_page_counter;
char buf[128];


#define MY_FILE "/mnt/host/mylog.txt"

void myPrintk (char* myMessage)
{
        mm_segment_t old_fs;

        struct file *file = filp_open(MY_FILE, O_RDWR | O_APPEND | O_CREAT, 0644);
        if (IS_ERR(file)) {
                printk("error occured while opening file %s, exiting...\n", MY_FILE);
                return;
        }

        old_fs = get_fs();
        set_fs(KERNEL_DS);
        file->f_op->write(file, myMessage, strlen(myMessage), &file->f_pos);
        set_fs(old_fs);
        filp_close(file, NULL);  
        file = NULL;
}

static inline int ext2s_add_nondir(struct dentry *dentry, struct inode *inode)
{
	int err = ext2s_add_link(dentry, inode);
	if (!err) {
		d_instantiate(dentry, inode);
		return 0;
	}
	inode_dec_link_count(inode);
	iput(inode);
	return err;
}

/*
 * Methods themselves.
 */

static struct dentry *ext2s_lookup(struct inode * dir, struct dentry *dentry, struct nameidata *nd)
{
	//printk(KERN_DEBUG "                                                                  ext2s_lookup & init gpc to 0.\n");
	//sprintf(buf,"%s\n", "ext2s_lookup & init gpc to 0.");
	//myPrintk(buf);
	ext2s_get_page_counter = 0;
	struct inode * inode;
	ino_t ino;
	
	if (dentry->d_name.len > EXT2_NAME_LEN) {
		//sprintf(buf,"%s %d\n", "ext2s_lookup leave counter", ext2s_get_page_counter);
		sprintf(buf," %d ", ext2s_get_page_counter);
		//myPrintk(buf);
		//printk(KERN_DEBUG "                                                                  ext2s_lookup leave counter %d\n", ext2s_get_page_counter);
		return ERR_PTR(-ENAMETOOLONG);
	}

	ino = ext2s_inode_by_name(dir, dentry);
	inode = NULL;
	if (ino) {
		inode = iget(dir->i_sb, ino);
		if (!inode) {
			//sprintf(buf,"%s %d\n", "ext2s_lookup leave counter", ext2s_get_page_counter);
			sprintf(buf," %d ", ext2s_get_page_counter);
			//myPrintk(buf);
			//printk(KERN_DEBUG "                                                                  ext2s_lookup leave counter %d\n", ext2s_get_page_counter);
			return ERR_PTR(-EACCES);
		}
	}
	//sprintf(buf,"%s %d\n", "ext2s_lookup leave counter", ext2s_get_page_counter);
	sprintf(buf," %d ", ext2s_get_page_counter);
	//myPrintk(buf);
	//printk(KERN_DEBUG "                                                                  ext2s_lookup leave counter %d\n", ext2s_get_page_counter);
	return d_splice_alias(inode, dentry);
}

struct dentry *ext2s_get_parent(struct dentry *child)
{
	unsigned long ino;
	struct dentry *parent;
	struct inode *inode;
	struct dentry dotdot;

	dotdot.d_name.name = "..";
	dotdot.d_name.len = 2;

	ino = ext2s_inode_by_name(child->d_inode, &dotdot);
	if (!ino)
		return ERR_PTR(-ENOENT);
	inode = iget(child->d_inode->i_sb, ino);

	if (!inode)
		return ERR_PTR(-EACCES);
	parent = d_alloc_anon(inode);
	if (!parent) {
		iput(inode);
		parent = ERR_PTR(-ENOMEM);
	}
	return parent;
} 

/*
 * By the time this is called, we already have created
 * the directory cache entry for the new file, but it
 * is so far negative - it has no inode.
 *
 * If the create succeeds, we fill in the inode information
 * with d_instantiate(). 
 */
static int ext2s_create (struct inode * dir, struct dentry * dentry, int mode, struct nameidata *nd)
{
	//printk(KERN_DEBUG "                                                                  ext2s_create & init gpc to 0.\n");		
	//sprintf(buf,"%s\n", "ext2s_create & init gpc to 0.");
	//myPrintk(buf);
	ext2s_get_page_counter = 0;
	struct inode * inode = ext2s_new_inode (dir, mode);
	int err = PTR_ERR(inode);
	if (!IS_ERR(inode)) {
		inode->i_op = &ext2s_file_inode_operations;
		if (ext2s_use_xip(inode->i_sb)) {
			inode->i_mapping->a_ops = &ext2s_aops_xip;
			inode->i_fop = &ext2s_xip_file_operations;
		} else if (test_opt(inode->i_sb, NOBH)) {
			inode->i_mapping->a_ops = &ext2s_nobh_aops;
			inode->i_fop = &ext2s_file_operations;
		} else {
			inode->i_mapping->a_ops = &ext2s_aops;
			inode->i_fop = &ext2s_file_operations;
		}
		mark_inode_dirty(inode);
		err = ext2s_add_nondir(dentry, inode);
	}
	//printk(KERN_DEBUG "                                                                  ext2s_create leave counter %d\n", ext2s_get_page_counter);
	//sprintf(buf,"%s %d\n", "ext2s_create leave counter", ext2s_get_page_counter);
	sprintf(buf," %d ", ext2s_get_page_counter);
	//myPrintk(buf);
	return err;
}

static int ext2s_mknod (struct inode * dir, struct dentry *dentry, int mode, dev_t rdev)
{
	struct inode * inode;
	int err;

	if (!new_valid_dev(rdev))
		return -EINVAL;

	inode = ext2s_new_inode (dir, mode);
	err = PTR_ERR(inode);
	if (!IS_ERR(inode)) {
		init_special_inode(inode, inode->i_mode, rdev);
#ifdef CONFIG_EXT2_FS_XATTR
		inode->i_op = &ext2s_special_inode_operations;
#endif
		mark_inode_dirty(inode);
		err = ext2s_add_nondir(dentry, inode);
	}
	return err;
}

static int ext2s_symlink (struct inode * dir, struct dentry * dentry,
	const char * symname)
{
	struct super_block * sb = dir->i_sb;
	int err = -ENAMETOOLONG;
	unsigned l = strlen(symname)+1;
	struct inode * inode;

	if (l > sb->s_blocksize)
		goto out;

	inode = ext2s_new_inode (dir, S_IFLNK | S_IRWXUGO);
	err = PTR_ERR(inode);
	if (IS_ERR(inode))
		goto out;

	if (l > sizeof (EXT2_I(inode)->i_data)) {
		/* slow symlink */
		inode->i_op = &ext2s_symlink_inode_operations;
		if (test_opt(inode->i_sb, NOBH))
			inode->i_mapping->a_ops = &ext2s_nobh_aops;
		else
			inode->i_mapping->a_ops = &ext2s_aops;
		err = page_symlink(inode, symname, l);
		if (err)
			goto out_fail;
	} else {
		/* fast symlink */
		inode->i_op = &ext2s_fast_symlink_inode_operations;
		memcpy((char*)(EXT2_I(inode)->i_data),symname,l);
		inode->i_size = l-1;
	}
	mark_inode_dirty(inode);

	err = ext2s_add_nondir(dentry, inode);
out:
	return err;

out_fail:
	inode_dec_link_count(inode);
	iput (inode);
	goto out;
}

static int ext2s_link (struct dentry * old_dentry, struct inode * dir,
	struct dentry *dentry)
{
	struct inode *inode = old_dentry->d_inode;

	if (inode->i_nlink >= EXT2_LINK_MAX)
		return -EMLINK;

	inode->i_ctime = CURRENT_TIME_SEC;
	inode_inc_link_count(inode);
	atomic_inc(&inode->i_count);

	return ext2s_add_nondir(dentry, inode);
}

static int ext2s_mkdir(struct inode * dir, struct dentry * dentry, int mode)
{
	struct inode * inode;
	int err = -EMLINK;

	if (dir->i_nlink >= EXT2_LINK_MAX)
		goto out;

	inode_inc_link_count(dir);

	inode = ext2s_new_inode (dir, S_IFDIR | mode);
	err = PTR_ERR(inode);
	if (IS_ERR(inode))
		goto out_dir;

	inode->i_op = &ext2s_dir_inode_operations;
	inode->i_fop = &ext2s_dir_operations;
	if (test_opt(inode->i_sb, NOBH))
		inode->i_mapping->a_ops = &ext2s_nobh_aops;
	else
		inode->i_mapping->a_ops = &ext2s_aops;

	inode_inc_link_count(inode);

	err = ext2s_make_empty(inode, dir);
	if (err)
		goto out_fail;

	err = ext2s_add_link(dentry, inode);
	if (err)
		goto out_fail;

	d_instantiate(dentry, inode);
out:
	return err;

out_fail:
	inode_dec_link_count(inode);
	inode_dec_link_count(inode);
	iput(inode);
out_dir:
	inode_dec_link_count(dir);
	goto out;
}

static int ext2s_unlink(struct inode * dir, struct dentry *dentry)
{
	struct inode * inode = dentry->d_inode;
	struct ext2s_dir_entry_2 * de;
	struct page * page;
	int err = -ENOENT;

	de = ext2s_find_entry (dir, dentry, &page);
	if (!de)
		goto out;

	err = ext2s_delete_entry (de, page);
	if (err)
		goto out;

	inode->i_ctime = dir->i_ctime;
	inode_dec_link_count(inode);
	err = 0;
out:
	return err;
}

static int ext2s_rmdir (struct inode * dir, struct dentry *dentry)
{
	struct inode * inode = dentry->d_inode;
	int err = -ENOTEMPTY;

	if (ext2s_empty_dir(inode)) {
		err = ext2s_unlink(dir, dentry);
		if (!err) {
			inode->i_size = 0;
			inode_dec_link_count(inode);
			inode_dec_link_count(dir);
		}
	}
	return err;
}

static int ext2s_rename (struct inode * old_dir, struct dentry * old_dentry,
	struct inode * new_dir,	struct dentry * new_dentry )
{
	struct inode * old_inode = old_dentry->d_inode;
	struct inode * new_inode = new_dentry->d_inode;
	struct page * dir_page = NULL;
	struct ext2s_dir_entry_2 * dir_de = NULL;
	struct page * old_page;
	struct ext2s_dir_entry_2 * old_de;
	int err = -ENOENT;

	old_de = ext2s_find_entry (old_dir, old_dentry, &old_page);
	if (!old_de)
		goto out;

	if (S_ISDIR(old_inode->i_mode)) {
		err = -EIO;
		dir_de = ext2s_dotdot(old_inode, &dir_page);
		if (!dir_de)
			goto out_old;
	}

	if (new_inode) {
		struct page *new_page;
		struct ext2s_dir_entry_2 *new_de;

		err = -ENOTEMPTY;
		if (dir_de && !ext2s_empty_dir (new_inode))
			goto out_dir;

		err = -ENOENT;
		new_de = ext2s_find_entry (new_dir, new_dentry, &new_page);
		if (!new_de)
			goto out_dir;
		inode_inc_link_count(old_inode);
		ext2s_set_link(new_dir, new_de, new_page, old_inode);
		new_inode->i_ctime = CURRENT_TIME_SEC;
		if (dir_de)
			drop_nlink(new_inode);
		inode_dec_link_count(new_inode);
	} else {
		if (dir_de) {
			err = -EMLINK;
			if (new_dir->i_nlink >= EXT2_LINK_MAX)
				goto out_dir;
		}
		inode_inc_link_count(old_inode);
		err = ext2s_add_link(new_dentry, old_inode);
		if (err) {
			inode_dec_link_count(old_inode);
			goto out_dir;
		}
		if (dir_de)
			inode_inc_link_count(new_dir);
	}

	/*
	 * Like most other Unix systems, set the ctime for inodes on a
 	 * rename.
	 * inode_dec_link_count() will mark the inode dirty.
	 */
	old_inode->i_ctime = CURRENT_TIME_SEC;

	ext2s_delete_entry (old_de, old_page);
	inode_dec_link_count(old_inode);

	if (dir_de) {
		ext2s_set_link(old_inode, dir_de, dir_page, new_dir);
		inode_dec_link_count(old_dir);
	}
	return 0;


out_dir:
	if (dir_de) {
		kunmap(dir_page);
		page_cache_release(dir_page);
	}
out_old:
	kunmap(old_page);
	page_cache_release(old_page);
out:
	return err;
}

const struct inode_operations ext2s_dir_inode_operations = {
	.create		= ext2s_create,
	.lookup		= ext2s_lookup,
	.link		= ext2s_link,
	.unlink		= ext2s_unlink,
	.symlink	= ext2s_symlink,
	.mkdir		= ext2s_mkdir,
	.rmdir		= ext2s_rmdir,
	.mknod		= ext2s_mknod,
	.rename		= ext2s_rename,
#ifdef CONFIG_EXT2_FS_XATTR
	.setxattr	= generic_setxattr,
	.getxattr	= generic_getxattr,
	.listxattr	= ext2s_listxattr,
	.removexattr	= generic_removexattr,
#endif
	.setattr	= ext2s_setattr,
	.permission	= ext2s_permission,
};

const struct inode_operations ext2s_special_inode_operations = {
#ifdef CONFIG_EXT2_FS_XATTR
	.setxattr	= generic_setxattr,
	.getxattr	= generic_getxattr,
	.listxattr	= ext2s_listxattr,
	.removexattr	= generic_removexattr,
#endif
	.setattr	= ext2s_setattr,
	.permission	= ext2s_permission,
};
